/* cdb_make_find.c: routines to search in in-progress cdb file */
/* This is a part of tinycdb package by Michael Tokarev, mjt@corpit.ru.
 * Public domain. */

#include <unistd.h>
#include "cdb_int.h"

/* XXX this routine uses pread() to avoid seeking */
static int
match(int fd, cdbi_t pos, const char *key, cdbi_t klen)
{
  unsigned char buf[64];
  if (pread(fd, buf, 4, pos) != 4)
    return -1;
  if (cdb_unpack(buf) != klen)
    return 0;

  pos += 8;
  while(klen > sizeof(buf)) {
    if (pread(fd, buf, sizeof(buf), pos) != sizeof(buf))
      return -1;
    if (memcmp(buf, key, sizeof(buf)) != 0)
      return 0;
    pos += sizeof(buf);
    key += sizeof(buf);
    klen -= sizeof(buf);
  }
  if (klen) {
    if (pread(fd, buf, klen, pos) != klen)
      return -1;
    if (memcmp(buf, key, klen) != 0)
      return 0;
  }
  return 1;
}

int
_cdb_make_find(struct cdb_make *cdbmp,
	       const void *key, cdbi_t klen, cdbi_t hval,
	       struct cdb_rl **rlp)
{
  struct cdb_rl *rl = cdbmp->cdb_rec[hval&255];
  int r, i;
  while(rl) {
    for(i = rl->cnt - 1; i >= 0; --i) { /* search backward */
      if (rl->rec[i].hval != hval)
	continue;
      if (cdbmp->cdb_bpos != cdbmp->cdb_buf) {
        if (write(cdbmp->cdb_fd, cdbmp->cdb_buf,
	          cdbmp->cdb_bpos - cdbmp->cdb_buf) < 0)
          return -1;
        cdbmp->cdb_bpos = cdbmp->cdb_buf;
      }
      r = match(cdbmp->cdb_fd, rl->rec[i].rpos, key, klen);
      if (!r)
	continue;
      if (r < 0)
	return -1;
      if (rlp)
	*rlp = rl;
      return i + 1;
    }
    rl = rl->next;
  }
  return 0;
}

int
cdb_make_exists(struct cdb_make *cdbmp,
                const void *key, cdbi_t klen)
{
  return _cdb_make_find(cdbmp, key, klen, cdb_hash(key, klen), NULL);
}
